#include "Compiler.h"
#include "codeGen/Quadruple.h"
#include "codeGen/RegisterAllocator.h"
#include <filesystem>
#include <iostream>
#include <cassert>

extern std::filesystem::path binaryPath;
const std::string Compiler::extension = "s";
void Compiler::externalCommand(std::filesystem::path lat) {
	auto obj = lat;
	obj.replace_extension("o");
	auto cmd = std::string("as --32 '") + lat.string() + "' -o '" + obj.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
	auto lib = binaryPath.parent_path() / "lib";
	auto runtime = lib / "runtime.o";
	auto libc = lib / "libc.a";
	auto result = obj;
	result.replace_extension("");
	cmd = std::string("ld -m elf_i386 -L'") + lib.string() + "' '" + runtime.string() + "' '" + libc.string() + "' '" + obj.string() + "' -o '" + result.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
}

string Compiler::compile(Visitable *v) {
	buf <<
		".globl _start\n"
	    ".globl exit\n"
		"_start:\n"
		"	pushl %ebp\n"
		"	movl %esp, %ebp\n"
		"	call main\n"
		"	pushl %eax\n"
		"	call exit\n";

	for (const auto& c : scope->classes) {
		c->calculateSize();
	}

	for (const auto& f : scope->functions) {
		compileFunction(f);
	}

	return buf.str();
}

void Compiler::compileFunction(const FunctionInfoPtr& f) {
	string name = mangleFunctionName(f);
	buf << ".globl \"" << name << "\"\n";

	if (!f->block) return;

	// TODO: jeśli funkcja jest klasowa, ustawić self na ... wartosc z ecx
	// c++ thiscall ecx, stack right-to-left (push c, push b, push a)

	buf << "\"" << name << "\":\n";
	stringstream asmGen;
	swap(buf, asmGen);
	exitLabel = make_shared<QLabel>("." + name + "_exit");
	QuadrupleGenerator::Result quadEnv = quadGen.compileFunction(f);
	try {
		RegisterAllocator regGen((quadEnv));
		regGen.allocate();
	} catch(const ParseError &e) {
		printFunction(quadEnv);
		throw ParseError(buf.str() + e.what());
	}

	printFunction(quadEnv);
	swap(buf, asmGen);
//	for (auto &b : quadEnv.blocks) {
//		for (auto &q : b->quads) {
//			q->generateAsm(*this);
//		}
//	}
	uint localsCount = temps.size() + localInfos.size();
	for (auto i : calleeSavedRegisters)
		append("PUSH", Register(i));
	buf <<
	    "\tpush %ebp\n"
	    "\tmov %esp, %ebp\n";
	if (localsCount) {
		buf << "\tsub $" + to_string(localsCount * 4) << ",%esp\n";
	}
	buf << asmGen.str();
	buf << "\"" << exitLabel->label << "\":\n" <<
		"MOVL %ebp, %esp\n"
	    "POP %ebp\n";
	for (auto i : calleeSavedRegisters)
		append("POP", Register(i));
	buf << "RET\n";

	for (auto &t : temps) t->localOffset = 0;
	for (auto &v : localInfos) v->localOffset = 0;
	temps.clear();
	localInfos.clear();
}

void Compiler::printFunction(QuadrupleGenerator::Result quadEnv) {
	int blkNo = 0;
	for (const auto& b : quadEnv.blocks) {
		cout << "----------------------------------------------------------------------------\n";
		cout << "blok " << blkNo << " " << b->getName() << endl;
		/*cout << "in: ";
		for (auto in : b->in) {
			cout << in->getName() << " ";
		}
		cout << "\nouts: ";
		for (auto out : b->out) {
			cout << out->getName() << " ";
		}
		cout << "\n";*/
		if (!b->flow.in.empty()) {
			cout << "in: ";
			for (const auto& v : b->flow.in) {
				cout << v->name << ", ";
			}
			cout << "\n----------\n";
		}
		if (!b->flow.def.empty()) {
			cout << "def: ";
			for (const auto& v : b->flow.def) {
				cout << v->name << ", ";
			}
			cout << "\n----------\n";
		}
		for (const auto& q : b->quads) {
			auto ret = q->toString();
			if (!ret.empty()) {
				auto len = 0;
				for (const auto &c : ret) {
					len++;
					if (c == '\t') len += 8 - (len % 8);
				}
				for (; len < 70; len++) ret += ' ';
				cout << ret << " alive: ";
				for (auto var : q->aliveAfter)
					cout << var->name << " ";
				cout << endl;

				cout << "\x1b[37;1m";
				stringstream asmStr;
				swap(asmStr, buf);
				q->generateAsm(*this);
				swap(asmStr, buf);
				buf << asmStr.str();
				cout << asmStr.str();
				cout << "\x1b[0m";
			}
		}
		if (!b->flow.use.empty()) {
			cout << "----------\nuse: ";
			for (const auto &v : b->flow.use) {
				cout << v->name << ", ";
			}
			cout << "\n";
		}
		if (!b->flow.out.empty()) {
			cout << "----------\nout: ";
			for (const auto &v : b->flow.out) {
				cout << v->name << ", ";
			}
			cout << "\n";
		}
		blkNo++;
	}
	generateStandardReturn();
}

const char* const Compiler::regs[6] = {"eax", "ebx", "ecx", "edx", "edi", "esi"};
const int Compiler::callerSavedRegisters[2] = {2, 3};
const int Compiler::calleeSavedRegisters[3] = {1, 4, 5};

string Compiler::getRef(VariablePtr &v) {
	if (v->registerColor == -1) {
		if (v->constExpr) {
			return "$" + to_string(v->val);
		}
		if (!v->localOffset) {
			if (v->info) {
				if (!v->info->localOffset) {
					localInfos.emplace(v->info);
					v->info->localOffset = 4 * (temps.size() + localInfos.size());
				}
				v->localOffset = v->info->localOffset;
			} else {
				temps.emplace(v);
				v->localOffset = 4 * (temps.size() + localInfos.size());
			}
		}
		return to_string(-v->localOffset) + "(%ebp)";
	} else {
		return Register(v);
	}
}

void Compiler::generateQAlloc(QAlloc &q) {
	// memory = extension
	throw runtime_error("Memory alloc unimplemented!");
}

void Compiler::generateQWrite(QWrite &q) {
	// memory = extension
	throw runtime_error("Memory write unimplemented!");
}

void Compiler::generateQAccess(QAccess &q) {
	// memory = extension
	throw runtime_error("Memory access unimplemented!");
}

void Compiler::generateQReturn(QReturn &q) {
	if (q.val)
		moveTo(q.val, 0);

	append("JMP", "\"" + exitLabel->label + "\"");
}

void Compiler::generateQCall(QCall &q) {
//	return Quadruple::generateAsm();
	if (q.self) {
		throw runtime_error("Class call unimplemented!");
	}
	append("CALL", "\"" + mangleFunctionName(q.fn) + "\"");
	if (q.params) {
		append("ADDL", "$" + to_string(q.params*4), "%esp");
	}
	auto popRegs = registersToSave(q);
	for (auto r = popRegs.rbegin(); r != popRegs.rend(); r++) {
		append("POPL", *r);
	}
	if (q.aliveAfter.count(q.loc)) {
		moveTo(0, q.loc);
	}
}

void Compiler::generateQParam(QParam &q) {
//	return Quadruple::generateAsm();
	auto reg = Register(q.param);
	if (q.num < 0) {
		auto offset = 4 * q.num;
		if (reg == 0) {
			q.param->localOffset = offset;
		} else {
			append("MOVL", to_string(-offset)+"(%ebp)", reg);
		}
	} else {
		if (q.num == 1 && q.call) {
			// first argument, need to save registers
			for (auto r : registersToSave(*q.call)) {
				append("PUSHL", r);
			}
		}
		append("PUSHL", getRef(q.param));
	}
}

void Compiler::generateQJump(QJump &q) {
//	return Quadruple::generateAsm();
	append("JMP", getBlockLabelText(q.target));
}

void Compiler::generateQJumpCond(QJumpCond &q) {
	auto op = q.op;
	if (q.left) {
		assert(!q.left->constExpr || !q.right->constExpr);
		auto regL = Register(q.left), regR = Register(q.right);
		auto locL = getRef(q.left);
		auto locR = getRef(q.right);
		if (regL == 0 && regR == 0) {
			append("MOVL", locR, regR);
			locR = regR;
		} else if (regL != 0 && regR == 0) {
			// left is in register but right is not - swap operands
			//  as right needs to be a register
			swap(locL, locR);
			if (op.kind() == Op::CMP && op != Op::EQ && op != Op::NEQ)
				op = op.neg();
		}
		append("CMP", locL, locR);
	} else {
		if (q.right->constExpr) {
			if ((op == Op::Not) ^ q.right->val) {
				generateQJump(q);
			}
			return;
		}
		auto tg = moveToAnyRegister(q.right);
		append("CMP", tg, tg);
	}
	if (op.kind() != Op::CMP && op.kind() != Op::UNARY) throw runtime_error("Unexpected operator in compare");
	string operand = getOpName(op);
	// getOpName returns SETZ, SETNE, SET*, replace it to JMP
	operand.replace(0,3, "J");
//	if (operand == "SETZ" || operand == "SETNZ") {
//
//	} else {
//		operand.replace(0,3, "JMP");
//	}
	append(operand, getBlockLabelText(q.target));
}

void Compiler::generateQAssign(QAssign &q) {
	if (!q.aliveAfter.count(q.loc)) return;
	Register tg(q.loc);
	if (q.args.size() == 1) {
		auto &arg = q.args[0];
		switch (q.op.op) {
			case Op::Not: {
				auto reg = Register(arg);
				if (reg == 0) {
					// avoids extra MOV when arg is already in register
					moveTo(arg, tg);
					reg = tg;
				}
				append("TEST", reg, reg);
				append("SETZ", tg);
				break;
			}
			case Op::Copy: {
				if (arg->constExpr) {
					append("MOVL", "$"+to_string(arg->val), getRef(q.loc));
					return;
				} else {
					moveTo(arg, tg);
				}
				break;
			}
			case Op::Neg: {
				moveTo(arg, tg);
				append("NEG", tg);
				break;
			}
			default: throw runtime_error("Unexpected operand");
		}
	} else {
		auto &i = q.args[0], &j = q.args[1];
		if (q.op == Op::Div || q.op == Op::Mod) {
			auto edx = Register("edx");
			bool edxAlive = false;
			for (auto &alive : q.aliveAfter)
				if (Register(alive) == edx) {
					edxAlive = true;
				}
			bool backupEdx = edx != tg && edx != Register(i) && (edx == Register(j) || edxAlive);
			if (backupEdx) append("MOVL", edx, "-4(%esp)");
			auto jLoc = edx != Register(j) ? getRef(j) : "-4(%esp)";
			moveTo(i, 0);
			append("CLTD"); // edx lost
			append("IDIVL", jLoc);
			append("MOVL", q.op == Op::Mod ? edx : 0, getRef(q.loc));
			// restore unnecessary if j was in edx and becomes dead and was NOT coalesced
			// not easy to check if not coalesced, so move anyway
			if (backupEdx && edxAlive) append("MOVL", "-4(%esp)", edx);
			return;
		} else if (q.op.kind() == Op::CMP) {
			auto loc = getRef(i);
			if (Register(i) == 0 && Register(j) == 0) {
				moveTo(i, 0);
				loc = "%eax";
			}

			append("CMPL", loc, getRef(j));
			append(q.op, tg);
		} else {
			assert(q.op.kind() == Op::BINARY);
			bool swapped = false;
			if (Register(j) == tg && tg != 0) {
				swapped = true;
				swap(i, j);
			}
			moveTo(i, tg);
			append(q.op, getRef(j), tg);
			if (swapped) {
				if (q.op == Op::Minus)
					append("NEG", tg);
			}
		}
	}
	if (tg == 0) {
		append("MOVL", tg, getRef(q.loc));
	}
}

void Compiler::generateQLabel(QLabel &q) {
	buf << "\"" << q.label << "\":" << endl;
}
